//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : Docking
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "docking.h"

//============================================================================//
void Docking::initialize()
//============================================================================//
{
    _Num_grid = max(_receptor->num_grid(),_ligand->num_grid());

    _fft_process  = new FFTProcess(_cputime,_parallel,_parameter);

    maxsize_voxel();

    return;
}

//============================================================================//
void Docking::maxsize_voxel()
//============================================================================//
{
    const int     num_fft = _Num_grid * 2;
    const int     natom = max(_receptor->num_atoms(),_ligand->num_atoms());
    const int     nag = natom * _Num_grid;
    const int     ng3 = _Num_grid * _Num_grid * _Num_grid;
    const int     nf3 = num_fft * num_fft * num_fft;

    const int     ns  = _parameter->_Num_sort;
    const int     nz  = _parameter->_NZangle;
    const int nproc2  = _parallel->nproc2();
    float     fmem;
    float     imem;

    fmem = nproc2*(ng3*3+natom*3+nag*3)
           +nproc2*2*nf3+2*nf3
           +nproc2*4*nf3
           +nz*ns;
    imem = nproc2*(ng3*2+natom*4)
           +4*nz*ns;

    double    mb = (fmem*4+imem*2)/1024/1024;
    mb += 10.0;       // margine
    printf("Memory requirement (/node)  = %.1f MB\n",mb); // This value is inaccurate. need revise
    _fft_process->alloc_array(num_fft);
    alloc_array(natom,nag,ng3);

    return;
}

//============================================================================//
void Docking::alloc_array(const int &natom,const int &nag,const int &ng3)
//============================================================================//
{
    const int nproc2    = _parallel->nproc2();

    _Grid_coord = new float[_Num_grid];
    if( !_Grid_coord ) {
        cerr << "[ERROR] Out of memory. Number of voxels for 1 axis = ("
             << _Num_grid
             << ") for (_Grid_coord) in docking.cpp!!\n";
        exit(1);
    }

    _Mol_coord = new float*[nproc2];
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _Mol_coord[i] = new float[natom*3];
        if( !_Mol_coord[i] ) {
            cerr << "[ERROR] Out of memory. Number of atoms times 3 axis = ("
                 << natom*3
                 << ") for (_Mol_coord) in docking.cpp!!\n";
            exit(1);
        }
    }

    _Memfw = ng3*3+natom*3+nag*3;
    _Fwork = new float[_Memfw*nproc2];
    if( !_Fwork ) {
        cerr << "[ERROR] Out of memory. Number of float arrays = ("
             << _Memfw*nproc2
             << ") for (_Fwork) in docking.cpp!!\n";
        exit(1);
    }

    _Memiw = ng3*2+natom*4;
    _Iwork = new int[_Memiw*nproc2];
    if( !_Iwork ) {
        cerr << "[ERROR] Out of memory. Number of int arrays = ("
             << _Memiw*nproc2
             << ") for (_Iwork) in docking.cpp!!\n";
        exit(1);
    }

    return;
}

//============================================================================//
void Docking::rec_init()
//============================================================================//
{
    const int ng1 = _Num_grid;
    const int ng2 = ng1 * ng1;
    const int ng3 = ng2 * ng1;
    float search_length   = _parameter->_Grid_size * _Num_grid;

    for( int i = 0 ; i < _Num_grid ; i++ ) {
        _Grid_coord[i] = search_length * ( -0.5 + ( 0.5 + i )/_Num_grid );
    }

    _receptor->shift_center(_Mol_coord[0]);
    create_voxel(_receptor,0);
    _fft_process->receptor_fft(_Fwork,&_Fwork[ng3]);

    return;
}

//============================================================================//
void Docking::create_voxel(Protein *rprotein,int myid2)
//============================================================================//
{

    if( _parameter->_Score_func == 1 ) {          // rPSC only
        rprotein->rpsc(_Num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
                       &_Fwork[_Memfw*myid2], _parameter->_rPSC_param_rec_core, _parameter->_rPSC_param_lig_core);

    } else if( _parameter->_Score_func == 2 ) { // rPSC + CHARMM19 electrostatic
        const float beta = -2800.0;         // MEGADOCK parameter
        rprotein->rpsc(_Num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
                       &_Fwork[_Memfw*myid2], _parameter->_rPSC_param_rec_core, _parameter->_rPSC_param_lig_core);

        rprotein->electro(beta,_parameter->_Elec_ratio,_Num_grid,_Grid_coord,_Mol_coord[myid2],
                          &_Iwork[_Memiw*myid2],&_Fwork[_Memfw*myid2]);

    } else if( _parameter->_Score_func == 3 ) {   // rPSC + CHARMM19 electrostatic + Receptor ACE
        const float beta = -2800.0;                 // MEGADOCK parameter
        rprotein->rpscace(_parameter->_ACE_ratio, _Num_grid, _Grid_coord, _Mol_coord[myid2],&_Iwork[_Memiw*myid2],
                          &_Fwork[_Memfw*myid2], _parameter->_rPSC_param_rec_core, _parameter->_rPSC_param_lig_core);
        rprotein->electro(beta,_parameter->_Elec_ratio,_Num_grid,_Grid_coord,_Mol_coord[myid2],
                          &_Iwork[_Memiw*myid2],&_Fwork[_Memfw*myid2]);
    }

    return;
}

//============================================================================//
void Docking::dockz()
//============================================================================//
{
    float theta[3];
    int xyz,myid2;

    const int ng1 = _Num_grid;
    const int ng2 = ng1 * ng1;
    const int ng3 = ng2 * ng1;
    const int nf1 = ng1 * 2;
    const int nf2 = nf1 * nf1;
    const int nf3 = nf2 * nf1;

    const int nc    = _parameter->_NZangle / 10;

    _fft_process->top_score_clean();

    cout << "\nLigand = " << _parameter->_LigPDB_file << endl;
    cout << "Target receptors:" << endl;;
    cout << " " << _parameter->_RecPDB_file << endl;
    cout << endl;
    cout.flush();

    #pragma omp parallel private(myid2,xyz,theta)
    {
#ifdef _OPENMP
        myid2 = omp_get_thread_num();
#else
        myid2 = 0;
#endif

        #pragma omp for
        for( int ang = 0 ; ang < _parameter->_NZangle ; ang++ ) {

            struct timeval et1, et2;

            if(nc!=0) {
                if( !((ang+1)%nc) ) {
                    printf("   >Ligand rotation = %5d / %5d (%2d)\n",ang+1,_parameter->_NZangle,myid2);
                }
            }

            if(myid2==0) gettimeofday(&et1,NULL); // ligand_voxelization --------------------

            _fft_process->rotation_index(myid2,ang); // _Rotation[myid2] = ang;

            for(xyz = 0 ; xyz < 3 ; xyz++ ) {
                theta[xyz] = _parameter->_Zangle[3*ang+xyz];
            }

            ligand_rotationz(theta,myid2);
            create_voxel(_ligand,myid2);

            if(myid2==0) {
                gettimeofday(&et2,NULL);
                _cputime->ligand_voxelization += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
            }

            if(myid2==0) gettimeofday(&et1,NULL); // ligand_fft -----------------------------
            _fft_process->ligand_fft(&_Fwork[_Memfw*myid2],&_Fwork[_Memfw*myid2+ng3],myid2);
            if(myid2==0) {
                gettimeofday(&et2,NULL);
                _cputime->fftprocess_ligand_fft += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
            }

            if(myid2==0) gettimeofday(&et1,NULL); // fft_inverse ----------------------------
            _fft_process->fft_score(myid2);
            if(myid2==0) {
                gettimeofday(&et2,NULL);
                _cputime->fftprocess_fft_inverse += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
            }

            if(myid2==0) gettimeofday(&et1,NULL); // score_sort -----------------------------
            _fft_process->score_sort(myid2);
            if(myid2==0) {
                gettimeofday(&et2,NULL);
                _cputime->fftprocess_score_sort += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
            }
        }
    }

    struct timeval et1, et2;

    gettimeofday(&et1,NULL);
    _fft_process->sort_index(_Fwork,_Iwork);
    gettimeofday(&et2,NULL);
    _cputime->fftprocess_sort_index += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

    return;
}

//============================================================================//
void Docking::ligand_rotationz(float *theta,int myid2)
//============================================================================//
{
    const int nl = _ligand->num_atoms();
    float     x, y, z;
    float     r11, r21, r31, r12, r22, r32, r13, r23, r33;

    r11 = cos(theta[0])*cos(theta[2])  -  sin(theta[0])*cos(theta[1])*sin(theta[2]);
    r21 = sin(theta[0])*cos(theta[2])  +  cos(theta[0])*cos(theta[1])*sin(theta[2]);
    r31 = sin(theta[1])*sin(theta[2]);

    r12 = -cos(theta[0])*sin(theta[2])  -  sin(theta[0])*cos(theta[1])*cos(theta[2]);
    r22 = -sin(theta[0])*sin(theta[2])  +  cos(theta[0])*cos(theta[1])*cos(theta[2]);
    r32 = sin(theta[1])*cos(theta[2]);

    r13 = sin(theta[0])*sin(theta[1]);
    r23 = -cos(theta[0])*sin(theta[1]);
    r33 = cos(theta[1]);

    for( int i = 0 ; i < nl ; i++ ) {
        x = _ligand->coordinate(i,0) - _ligand->_Center[0];
        y = _ligand->coordinate(i,1) - _ligand->_Center[1];
        z = _ligand->coordinate(i,2) - _ligand->_Center[2];
        _Mol_coord[myid2][3*i  ] = r11 * x + r12 * y + r13 * z;
        _Mol_coord[myid2][3*i+1] = r21 * x + r22 * y + r23 * z;
        _Mol_coord[myid2][3*i+2] = r31 * x + r32 * y + r33 * z;
    }

    return;
}

//============================================================================//
void Docking::output()
//============================================================================//
{
    char      buff[256];
    float     coord[3], coordl[3], theta[3];

    const int nf1 = _Num_grid * 2;
    const float   grid_width     = _parameter->_Grid_size;
    const int num_output  = _parameter->_Num_output;

    ofstream Output(_parameter->_RLOut_file.c_str()); // open Output file
    if( !Output ) {
        cerr << "[ERROR] Output file [" << _parameter->_RLOut_file << "] not open!!" << endl;
        exit(1);
    }

    sprintf(buff,"%d\t%.2f",nf1,grid_width); // Line 1 : Number of FFT, Grid width
    Output << buff << endl;

    for( int xyz = 0 ; xyz < 3 ; xyz++ ) {
        theta[xyz] = _receptor->angle(xyz);
    }

    sprintf(buff,"\t%f\t%f\t%f",theta[0],theta[1],theta[2]); // Line 2 : angle of Receptor
    Output << buff << endl;

    for( int xyz = 0 ; xyz < 3 ; xyz++ ) {
        coord[xyz] = _receptor->center(xyz);
        coordl[xyz] = _ligand->center(xyz);
    }

    sprintf(buff,"\t%f\t%f\t%f",coord[0],coord[1],coord[2]); // Line 3 : Receptor PDB file, center coordinate
    Output << _parameter->_RecPDB_file << buff << endl;

    sprintf(buff,"\t%f\t%f\t%f",coordl[0],coordl[1],coordl[2]); // Line 4 : Ligand PDB file, center coordinate
    Output << _parameter->_LigPDB_file << buff << endl;

    for( int rank = 0 ; rank < num_output ; rank++ ) {
        const int n = _fft_process->top_index(rank,0);
        for( int k = 0 ; k < 3 ; k++ ) {
            theta[k] = _parameter->_Zangle[3*n+k];
            if( theta[k] >= PI ) {
                theta[k] -= PI*2.0;
            }
        }
        sprintf(buff,"%f\t%f\t%f\t%d\t%d\t%d\t%.2f",theta[0],theta[1],theta[2], // Line 5- : angle, coord, score
                _fft_process->top_index(rank,1),_fft_process->top_index(rank,2),
                _fft_process->top_index(rank,3),_fft_process->top_score(rank));
        Output << buff << endl;
    }

    Output.close(); // close Output file

    return;
}

//============================================================================//
void Docking::output_detail()
//============================================================================//
{

// ##############################
// #Output detailed result      #
// ##############################

    float score_rpsc, score_elec, score_ace, score_total;
    float score_rpsc_plus, score_rpsc_minus;
    float score_elec_plus, score_elec_minus;
    float score_ace_plus, score_ace_minus;
    int count_rpsc_plus, count_rpsc_minus;
    int count_elec_plus, count_elec_minus;
    int count_ace_plus, count_ace_minus;
    float theta[3], coord[3], coordl[3], diff_dis[3];
    float *rec_r, *rec_i, *rec_r_rpsc, *lig_r, *lig_i;
    int myid2, la, lb, ma, mb, na, nb;
    int cor_coord[3], tem_coord[3];
    char buff[256];

    const int ng1 = _Num_grid;
    const int ng2 = ng1 * ng1;
    const int ng3 = ng2 * ng1;
    const int nf1 = ng1 * 2;
    const int nf2 = nf1 * nf1;
    const int nf3 = nf2 * nf1;

    const float grid_width = _parameter->_Grid_size;
    const int num_output = _parameter->_Num_output;
    const int orig_ACE_ratio = _parameter->_ACE_ratio;

    rec_r = new float[ng3];
    rec_r_rpsc = new float[ng3];
    rec_i = new float[ng3];
    lig_r = new float[ng3];
    lig_i = new float[ng3];

    ofstream Output_detail(_parameter->_RLOut_file_detail.c_str()); // open Output file
    if( !Output_detail ) {
        cerr << "[ERROR] Output_detail file [" << _parameter->_RLOut_file_detail << "] not open!!" << endl;
        exit(1);
    }

    sprintf(buff,"%d\t%.2f",nf1,grid_width); // Line 1 : Number of FFT, Grid width
    Output_detail << buff << endl;

    for( int xyz = 0 ; xyz < 3 ; xyz++ ) {
        theta[xyz] = _receptor->angle(xyz);
    }

    sprintf(buff,"\t%f\t%f\t%f",theta[0],theta[1],theta[2]); // Line 2 : angle of Receptor
    Output_detail << buff << endl;

    for( int xyz = 0 ; xyz < 3 ; xyz++ ) {
        coord[xyz] = _receptor->center(xyz);
        coordl[xyz] = _ligand->center(xyz);
    }

    sprintf(buff,"\t%f\t%f\t%f",coord[0],coord[1],coord[2]); // Line 3 : Receptor PDB file, center coordinate
    Output_detail << _parameter->_RecPDB_file << buff << endl;

    sprintf(buff,"\t%f\t%f\t%f",coordl[0],coordl[1],coordl[2]); // Line 4 : Ligand PDB file, center coordinate
    Output_detail << _parameter->_LigPDB_file << buff << endl;

    sprintf(buff,"\n *** score around default position (and each +-1 position)*** \n");
    Output_detail << buff << endl;

    for(int i=0; i<3; i++) { // calculate correct position
        diff_dis[i] = coord[i] - coordl[i];
        cor_coord[i] =round(diff_dis[i] / grid_width);
        if(cor_coord[i] < 0) cor_coord[i] += nf1;
    }
    sprintf(buff,"%f\t%f\t%f\t%d\t%d\t%d\t(correct position)\n",theta[0],theta[1],theta[2],cor_coord[0],cor_coord[1],cor_coord[2]);
    Output_detail << buff << endl;

    sprintf(buff," \t \t \t \t \t \t \t \t \ttotal\trPSC\tELEC\tRecACE");
    Output_detail << buff << endl;

    // Prepare voxel scores -------------------------------------------------
    _receptor->shift_center(_Mol_coord[0]);
    create_voxel(_receptor,0); // Receptor --------
    for (int i=0; i<ng3; i++) {
        rec_r[i] = _Fwork[i];
        rec_i[i] = _Fwork[i+ng3];
    }

    _parameter->_ACE_ratio = 0.0; // Receptor(except ACE) --------
    create_voxel(_receptor,0);

    for (int i=0; i<ng3; i++) {
        rec_r_rpsc[i] = _Fwork[i];
    }

    for(int j = 0 ; j < 3 ; j++ ) { // Ligand --------
        theta[j] = 0.0;
    }
    ligand_rotationz(theta,0);
    create_voxel(_ligand,0);
    for (int i=0; i<ng3; i++) {
        lig_r[i] = _Fwork[i];
        lig_i[i] = _Fwork[i+ng3];
    }

    for(int around=0; around<7; around++) {
        score_rpsc = 0.0;
        score_elec = 0.0;
        score_ace = 0.0;
        score_total = 0.0;
        if(around==1) cor_coord[0]--;
        if(around==2) cor_coord[0]=cor_coord[0]+2;
        if(around==3) {
            cor_coord[0]--;
            cor_coord[1]--;
        }
        if(around==4) cor_coord[1]=cor_coord[1]+2;
        if(around==5) {
            cor_coord[1]--;
            cor_coord[2]--;
        }
        if(around==6) cor_coord[2]=cor_coord[2]+2;
        //printf("\n %d, %d, %d \n",cor_coord[0],cor_coord[1],cor_coord[2]);

        for( int xyz = 0 ; xyz < 3 ; xyz++ ) {
            tem_coord[xyz] = (3*ng1-cor_coord[xyz]) % (2*ng1);
        }

        if (tem_coord[0]<ng1) {
            la=ng1-tem_coord[0];
            lb=ng1;
        } else {
            la=0;
            lb=2*ng1-tem_coord[0];
        }
        if (tem_coord[1]<ng1) {
            ma=ng1-tem_coord[1];
            mb=ng1;
        } else {
            ma=0;
            mb=2*ng1-tem_coord[1];
        }
        if (tem_coord[2]<ng1) {
            na=ng1-tem_coord[2];
            nb=ng1;
        } else {
            na=0;
            nb=2*ng1-tem_coord[2];
        }
        for (int l=la; l < lb; l++) {
            const int rxc = ng2 * (l+tem_coord[0]-ng1);
            const int lxc = ng2 * l;
            for (int m=ma; m < mb; m++) {
                const int ryc = rxc + ng1 * (m+tem_coord[1]-ng1);
                const int lyc = lxc + ng1 * m;
                for (int n=na; n < nb; n++) {
                    const int rzc=ryc+(tem_coord[2]+n-ng1);
                    const int lzc=lyc+n;
                    score_rpsc += rec_r_rpsc[rzc]*lig_r[lzc];
                    score_elec += rec_i[rzc]*lig_i[lzc];
                    score_ace += (rec_r[rzc] - rec_r_rpsc[rzc])*lig_r[lzc];
                }
            }
        }
        score_total = score_rpsc + score_elec + score_ace;
        sprintf(buff,"%f\t%f\t%f\t%d\t%d\t%d\t%7.2f\t%7.2f\t%7.2f\t%7.2f"
                ,theta[0],theta[1],theta[2],cor_coord[0],cor_coord[1],cor_coord[2]
                ,score_total,score_rpsc,score_elec,score_ace);
        Output_detail << buff << endl;
    }

    // Line 19- : docking results
    sprintf(buff,"\n *** docking results *** \n");
    Output_detail << buff;
    sprintf(buff," \t \t \t \t \t \t \t \t \ttotal\trPSC\tELEC\tRecACE");
    Output_detail << buff << endl;

    for( int rank = 0 ; rank < num_output ; rank++ ) {
        const int angle_num = _fft_process->top_index(rank,0); // prepare angle & coord
        for( int xyz = 0 ; xyz < 3 ; xyz++ ) {
            theta[xyz] = _parameter->_Zangle[3*angle_num+xyz];
            cor_coord[xyz] = _fft_process->top_index(rank,xyz + 1);
            tem_coord[xyz] = (3*ng1-cor_coord[xyz]) % (2*ng1);
        }

        ligand_rotationz(theta,0);
        create_voxel(_ligand,0);

        if (tem_coord[0]<ng1) {
            la=ng1-tem_coord[0];
            lb=ng1;
        } else {
            la=0;
            lb=2*ng1-tem_coord[0];
        }
        if (tem_coord[1]<ng1) {
            ma=ng1-tem_coord[1];
            mb=ng1;
        } else {
            ma=0;
            mb=2*ng1-tem_coord[1];
        }
        if (tem_coord[2]<ng1) {
            na=ng1-tem_coord[2];
            nb=ng1;
        } else {
            na=0;
            nb=2*ng1-tem_coord[2];
        }
        const int tem_num_grid3 = (lb-la)*(mb-ma)*(nb-na);

        score_rpsc = 0.0;
        score_elec = 0.0;
        score_ace = 0.0;
        score_total = 0.0;

        score_rpsc_plus = 0.0;
        score_rpsc_minus = 0.0;
        score_elec_plus = 0.0;
        score_elec_minus = 0.0;
        score_ace_plus = 0.0;
        score_ace_minus = 0.0;

        count_rpsc_plus = 0;
        count_rpsc_minus = 0;
        count_elec_plus = 0;
        count_elec_minus = 0;
        count_ace_plus = 0;
        count_ace_minus = 0;

        for (int l=la; l < lb; l++) {
            const int rxc = ng2 * (l+tem_coord[0]-ng1);
            const int lxc = ng2 * l;
            for (int m=ma; m < mb; m++) {
                const int ryc = rxc + ng1 * (m+tem_coord[1]-ng1);
                const int lyc = lxc + ng1 * m;
                for (int n=na, rzc=ryc+(tem_coord[2]+na-ng1), lzc=lyc+na, lzc_i=lyc+na+ng3; n < nb; n++, rzc++, lzc++, lzc_i++) { // rzc & lzc increment for speed
                    const float rpsc = rec_r_rpsc[rzc]*_Fwork[lzc];
                    const float elec = rec_i[rzc]*_Fwork[lzc_i];
                    const float ace = (rec_r[rzc] - rec_r_rpsc[rzc])*_Fwork[lzc];
                    if (rpsc < 0) {
                        score_rpsc_minus += rpsc;
                        count_rpsc_minus++;
                    } else if(rpsc > 0) {
                        score_rpsc_plus += rpsc;
                        count_rpsc_plus++;
                    }
                    if (elec < 0) {
                        score_elec_minus += elec;
                        count_elec_minus++;
                    } else if(elec > 0) {
                        score_elec_plus += elec;
                        count_elec_plus++;
                    }
                    if (ace < 0) {
                        score_ace_minus += ace;
                        count_ace_minus++;
                    } else if(ace > 0) {
                        score_ace_plus += ace;
                        count_ace_plus++;
                    }

                }
            }
        }
        score_rpsc =score_rpsc_plus + score_rpsc_minus;
        score_elec =score_elec_plus + score_elec_minus;
        score_ace = score_ace_plus + score_ace_minus;

        score_total = score_rpsc + score_elec + score_ace;

        for( int xyz = 0 ; xyz < 3 ; xyz++ ) { // PI wo koeru theta ha nai kiga suru
            if( theta[xyz] >= PI ) {
                theta[xyz] -= PI*2.0;
            }
        }

        sprintf(buff,"%f\t%f\t%f\t%d\t%d\t%d\t%7.2f\t%7.2f\t%7.2f\t%7.2f"
                ,theta[0],theta[1],theta[2],cor_coord[0],cor_coord[1],cor_coord[2]
                ,score_total,score_rpsc,score_elec,score_ace);
        Output_detail << buff << endl;

    }

    Output_detail.close(); // close Output file

    return;
}
